"""Sphinx configuration."""
from datetime import datetime


project = "Language models for astrochemistry"
author = "Kelvin Lee"
copyright = f"{datetime.now().year}, {author}"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_click",
    "sphinx_rtd_theme",
]
autodoc_typehints = "description"
html_theme = "sphinx_rtd_theme"



"""
layers.py

This module is intended for designing the building
blocks of models. These blocks are then used to
compose models in `astrochem_embedding.models`.
"""

import torch
from torch import nn
import pytorch_lightning as pl



import selfies as sf
import numpy as np
from tqdm.auto import tqdm
from ruamel.yaml import YAML
from joblib import Parallel, delayed

from astrochem_embedding import get_paths, Translator


NUM_JOBS = 24


def get_label(selfie: str, translator, index: int, output_array: np.ndarray):
    label, _ = translator.tokenize(selfie)
    output_array[index, :] = label


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


paths = get_paths()

with open(paths.get("processed").joinpath("selfies.txt")) as read_file:
    selfs = list(map(lambda x: x.strip(), read_file.readlines()))
# get the max length for the entire dataset
lengths = Parallel(n_jobs=NUM_JOBS)(delayed(sf.len_selfies)(s) for s in tqdm(selfs))
max_length = max(lengths)
alphabet = sf.get_alphabet_from_selfies(selfs)
alphabet.add("[nop]")
alphabet.add("[unk]")
alphabet = list(sorted(alphabet))

translator = Translator(alphabet, max_length)
print(translator)
translator.to_yaml(paths.get("processed").joinpath("translator.yml"))

vocab_mapping = {s: i for i, s in enumerate(alphabet)}

labels = np.memmap(
    paths.get("processed").joinpath("labels.npy"),
    dtype=np.uint16,
    shape=(len(selfs), max_length),
    mode="w+",
)

Parallel(n_jobs=NUM_JOBS)(
    delayed(get_label)(s, translator, i, labels) for i, s in tqdm(enumerate(selfs))
)
blah = labels.copy()
labels._mmap.close()

np.save(paths.get("processed").joinpath("labels.npy"), blah)



"""Language models for astrochemistry."""

from astrochem_embedding.utils import get_paths, Translator, get_pretrained_path
from astrochem_embedding import models, pipeline, layers

from astrochem_embedding.models.models import VICGAE, GRUAutoEncoder
from astrochem_embedding.pipeline.data import MaskedStringDataModule, StringDataModule

__version__ = "0.2.0"



"""
models.py

This module is intended for *composed* models; i.e.
ready to training/usage, based off of layers defined in
either `torch`, other packages, or in `astrochem_embedding.layers`.
"""

from typing import Union, Iterable
from pathlib import Path

import torch
import joblib
from torch import Tensor  # this is used for type annotations
from torch import nn
from torch.nn import functional as F
import wandb
import pytorch_lightning as pl

from astrochem_embedding.models import layers
from astrochem_embedding import get_paths, get_pretrained_path, Translator


class AutoEncoder(pl.LightningModule):
    def __init__(
        self,
        embedding_dim: int,
        encoder: nn.Module,
        decoder: nn.Module,
        lr: float = 1e-3,
        vocab_yaml: Union[str, None] = None,
    ):
        super().__init__()
        if not vocab_yaml:
            paths = get_paths()
            vocab_yaml = get_pretrained_path().joinpath("translator.yml")
        self.vocab = Translator.from_yaml(vocab_yaml)
        vocab_size = len(self.vocab)
        self.embedding = nn.Embedding(vocab_size, embedding_dim)
        self.encoder = encoder
        self.decoder = decoder
        self.metric = nn.BCELoss()
        self.example_input_array = torch.randint(0, vocab_size, size=(64, 10))

    @property
    def vocab_size(self) -> int:
        return len(self.vocab)

    def tokens2smiles(self, tokens: Iterable[int]):
        return self.vocab.indices_to_smiles(tokens)

    @torch.no_grad()
    def embed_molecule(self, X: torch.Tensor) -> torch.Tensor:
        if X.ndim == 1:
            X.unsqueeze_(0)
        # shape [N, S, D] for S sequence length and D features
        word_embeddings = self.embedding(X)
        # extract recursive embeddings
        z_o, z_h = self.encoder(word_embeddings)
        mask = X != self.vocab.alphabet.index("[nop]")
        # perform a normalized summation over non-empty tokens
        z = (
            torch.einsum("ijk,ij->ik", z_o, mask.float())
            / self.hparams.embedding_dim
        )
        return z

    @torch.no_grad()
    def embed_smiles(self, smiles: str) -> torch.Tensor:
        labels, _ = self.vocab.tokenize_smiles(smiles)
        tokens = torch.LongTensor(labels)
        return self.embed_molecule(tokens.unsqueeze(0))

    def configure_optimizers(self):
        opt = torch.optim.Adam(self.parameters(), lr=self.hparams.lr)
        return opt

    def step(self, batch, prefix: str):
        X1, X2, Y = batch
        targets = F.one_hot(Y, num_classes=self.vocab_size)
        output = self(X1)
        loss = self.metric(output, targets.float())
        self.log("{prefix}_loss", loss)
        return loss

    def training_step(self, batch, batch_idx):
        loss = self.step(batch, "train")
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.step(batch, "validation")
        # get some examples
        # ex_targets = labels[:10]
        # with torch.no_grad():
        #    ex_outputs = self(ex_targets)
        # ex_targets = ex_targets.cpu().numpy()
        # ex_outputs = ex_outputs.argmax(dim=-1).cpu().numpy()
        # target_smiles = [self.tokens2smiles(t) for t in ex_targets]
        # output_smiles = [self.tokens2smiles(t) for t in ex_outputs]
        # self.log("example_smiles",
        #    {"targets": target_smiles, "outputs": output_smiles}
        # )
        return loss

    @classmethod
    def from_pretrained(cls):
        pretrained_path = get_pretrained_path().joinpath(f"{cls.__name__}.pkl")
        return joblib.load(pretrained_path)


class GRUAutoEncoder(AutoEncoder):
    def __init__(
        self,
        embedding_dim: int,
        z_dim: int,
        num_layers: int,
        dropout: float = 0.0,
        lr: float = 1e-3,
        vocab_yaml: Union[str, None] = None,
    ):
        if not vocab_yaml:
            paths = get_paths()
            vocab_yaml = get_pretrained_path().joinpath("translator.yml")
        translator = Translator.from_yaml(vocab_yaml)
        vocab_size = len(translator)
        encoder = nn.GRU(
            embedding_dim,
            z_dim,
            num_layers=num_layers,
            dropout=dropout,
            batch_first=True,
        )
        decoder = nn.GRU(
            z_dim, z_dim, num_layers=num_layers, dropout=dropout, batch_first=True
        )
        output = nn.Sequential(nn.Linear(z_dim, vocab_size), nn.Softmax(dim=-1))
        super().__init__(embedding_dim, encoder, decoder, lr, vocab_yaml)
        self.output = output
        self.save_hyperparameters()

    def forward(self, X):
        embeddings = self.embedding(X)
        z_o, z_h = self.encoder(embeddings)
        o_o, o_h = self.decoder(z_o, z_h)
        output = self.output(o_o)
        return output


class VICGAE(GRUAutoEncoder):
    def __init__(
        self,
        embedding_dim: int,
        z_dim: int,
        num_layers: int,
        dropout: float = 0.0,
        lr: float = 1e-3,
        vocab_yaml: Union[str, None] = None,
    ):
        super().__init__(embedding_dim, z_dim, num_layers, dropout, lr, vocab_yaml)
        self.vic_reg = layers.VICRegularization()

    def _vic_regularization(self, batch):
        X1, X2, Y = batch
        mask_1 = X1 != self.vocab.alphabet.index("[nop]")
        mask_2 = X2 != self.vocab.alphabet.index("[nop]")
        # shape [N, S, Z] for S sequence length
        word_z_1 = self.embedding(X1)
        word_z_2 = self.embedding(X2)
        # idea behind summmation is to have molecule = sum of characters
        z_1 = (
            torch.einsum("ijk,ij->ik", word_z_1, mask_1.float())
            / self.hparams.embedding_dim
        )
        z_2 = (
            torch.einsum("ijk,ij->ik", word_z_2, mask_2.float())
            / self.hparams.embedding_dim
        )
        v, i, c = self.vic_reg(z_1, z_2)
        return v, i, c

    def step(self, batch, prefix: str):
        X1, X2, Y = batch
        targets = F.one_hot(Y, num_classes=self.vocab_size).float()
        output_1 = self(X1)
        output_2 = self(X2)
        # include the VIC regularization
        v, i, c = self._vic_regularization(batch)
        loss_1 = self.metric(output_1, targets)
        loss_2 = self.metric(output_2, targets)
        loss = loss_1 + loss_2 + v + i + c
        self.log(f"{prefix}_loss", {"loss": loss, "v": v, "i": i, "c": c})
        return loss



"""
layers.py

This module is intended for storing the building blocks
of models.
"""

import torch
from torch import Tensor  # this is used for type annotations
from torch import nn
from torch.nn import functional as F
import pytorch_lightning as pl

from astrochem_embedding import layers


class VarianceHinge(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, embedding):
        # variance throughout the batch for each embedding dimension
        variance = 1.0 - torch.sqrt(embedding.var(dim=0) + 1e-10)
        # average along embedding dimensions
        return torch.maximum(torch.zeros_like(variance), variance).mean()


class CovarianceLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, embedding) -> float:
        # covariance along embedding dimensions
        covariance = embedding.T.cov()
        mask = ~torch.eye(*covariance.shape, device=covariance.device, dtype=bool)
        # mask out the diagonal elements, we just want to minimize covariance
        loss = (covariance[mask] ** 2.0).mean()
        return loss


class VICRegularization(nn.Module):
    def __init__(self):
        super().__init__()
        self.variance = VarianceHinge()
        self.covariance = CovarianceLoss()
        self.invariance = nn.MSELoss()

    def forward(self, z_1: torch.Tensor, z_2: torch.Tensor):
        packed = [z_1, z_2]
        variance = sum([self.variance(z) for z in packed])
        covariance = sum([self.covariance(z) for z in packed])
        invariance = self.invariance(z_1, z_2)
        return variance, invariance, covariance



import torch
import pytorch_lightning as pl

from astrochem_embedding.pipeline.data import (
    SELFIESData,
    StringDataModule,
    MaskedStringDataModule,
)
from astrochem_embedding import get_paths
from astrochem_embedding.models import models

pl.seed_everything(215015)

BATCH_SIZE = 128
NUM_WORKERS = 12
EMBEDDING_DIM = 128
Z_DIM = 32
NUM_LAYERS = 1
LR = 1e-4

model = models.VICGAE(EMBEDDING_DIM, Z_DIM, NUM_LAYERS, lr=LR)

data = MaskedStringDataModule(BATCH_SIZE, NUM_WORKERS)

logger = pl.loggers.TensorBoardLogger(
    "tb_logs", name="VICAstrochemEmbedder", log_graph=True
)
summarizer = pl.callbacks.ModelSummary(max_depth=-1)

trainer = pl.Trainer(max_epochs=5, callbacks=[summarizer], gpus=1, logger=logger)
trainer.fit(model, data)

paths = get_paths()
trainer.save_checkpoint(paths.get("models").joinpath("VICGAE.ckpt"))



import torch

from astrochem_embedding.models import models


@torch.no_grad()
def test_gruautoencoder():
    model = models.GRUAutoEncoder(embedding_dim=8, z_dim=16, num_layers=1)
    # expect a tensor of shape batch, seq_length
    test_array = torch.randint(0, 10, size=(32, 10))
    output = model(test_array)


@torch.no_grad()
def test_gruautoencoder_multi():
    model = models.GRUAutoEncoder(embedding_dim=8, z_dim=16, num_layers=3)
    # expect a tensor of shape batch, seq_length
    test_array = torch.randint(0, 10, size=(32, 10))
    output = model(test_array)


@torch.no_grad()
def test_gruautoencoder_step():
    model = models.GRUAutoEncoder(embedding_dim=8, z_dim=16, num_layers=1)
    # expect a tensor of shape batch, seq_length
    test_array = torch.randint(0, 10, size=(32, 10))
    model.step((test_array, test_array, test_array), "test")


@torch.no_grad()
def test_vicgae():
    model = models.VICGAE(embedding_dim=8, z_dim=16, num_layers=1)
    # expect a tensor of shape batch, seq_length
    test_array = torch.randint(0, 10, size=(32, 10))
    output = model(test_array)


@torch.no_grad()
def test_vicgae_multi():
    model = models.VICGAE(embedding_dim=8, z_dim=16, num_layers=3)
    # expect a tensor of shape batch, seq_length
    test_array = torch.randint(0, 10, size=(32, 10))
    output = model(test_array)


@torch.no_grad()
def test_vicgae_step():
    model = models.VICGAE(embedding_dim=8, z_dim=16, num_layers=1)
    # expect a tensor of shape batch, seq_length
    test_array = torch.randint(0, 10, size=(32, 10))
    model.step((test_array, test_array, test_array), "test")


#def test_load_gruautoencoder():
#    model = models.GRUAutoEncoder.from_pretrained()


def test_load_vicgae():
    model = models.VICGAE.from_pretrained()



"""
data.py


This module is intended for dataset abstractions,
comprising data generation, I/O, and `DataModule`
and `Dataset` definitions.
"""
from abc import abstractmethod
from typing import Union, Type
from os import environ
from warnings import warn

import torch
import numpy as np
import selfies as sf
from torch.utils.data import DataLoader, Dataset, random_split
import pytorch_lightning as pl
from sklearn.preprocessing import OneHotEncoder

from astrochem_embedding import get_paths, Translator


class SELFIESDataset(Dataset):
    def __init__(self, path=None):
        super().__init__()
        if not path:
            paths = get_paths()
            path = paths.get("processed").joinpath("labels.npy")
        temp = np.load(path)
        shape, dtype = temp.shape, temp.dtype
        del temp
        data = np.memmap(path, dtype=dtype, shape=shape)
        self.data = data
        self.shape = shape
        self.encoder = OneHotEncoder(sparse=False)

    def __getitem__(self, index):
        return self.data[index].astype(np.int)

    def __len__(self):
        return self.shape[0]


class StringDataset(Dataset):
    def __init__(self, path=None):
        super().__init__()
        if not path:
            paths = get_paths()
            path = paths.get("processed").joinpath("selfies.txt")
        with open(path, "r") as read_file:
            self.data = read_file.readlines()
            self.data = [s.strip() for s in self.data]
        self.translator = Translator.from_yaml(
            paths.get("processed").joinpath("translator.yml")
        )

    def __getitem__(self, index) -> torch.Tensor:
        label, _ = self.translator.tokenize(self.data[index])
        label = torch.LongTensor(label)
        return label, label, label

    def __len__(self) -> int:
        return len(self.data)


class MaskedStringDataset(StringDataset):
    def __init__(self, path=None):
        super().__init__(path)
        self.nop = self.translator.alphabet.index("[nop]")
        self.unk = self.translator.alphabet.index("[unk]")

    def __getitem__(self, index) -> torch.Tensor:
        target, _ = self.translator.tokenize(self.data[index])
        input_1, input_2 = target.copy(), target.copy()
        actual_length = len(list(filter(lambda x: x != self.nop, input_1)))
        # todo: option for multiple tokens to blanked
        indices = np.arange(actual_length)
        index_1 = np.random.choice(indices)
        index_2 = np.random.choice(indices)
        # replace token with unknown mask
        input_1[index_1] = self.unk
        input_2[index_2] = self.unk
        return (
            torch.LongTensor(input_1),
            torch.LongTensor(input_2),
            torch.LongTensor(target),
        )


class SELFIESData(pl.LightningDataModule):
    def __init__(self, batch_size: int = 256, num_workers: int = 0, path=None):
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.dataset = SELFIESDataset(path)

    def setup(self, stage=None):
        length = len(self.dataset)
        val_size = int(length * 0.2)
        train_size = length - val_size
        self.train, self.val = random_split(self.dataset, [train_size, val_size])

    @property
    def vocab_size(self) -> int:
        return self.dataset.max() + 1

    def train_dataloader(self):
        return DataLoader(
            self.train, batch_size=self.batch_size, num_workers=self.num_workers
        )

    def val_dataloader(self):
        return DataLoader(
            self.val, batch_size=self.batch_size, num_workers=self.num_workers
        )


class StringDataModule(SELFIESData):
    def __init__(self, batch_size: int = 256, num_workers: int = 0, path=None):
        super().__init__(batch_size, num_workers, path)
        self.dataset = StringDataset(path)


class MaskedStringDataModule(SELFIESData):
    def __init__(self, batch_size: int = 256, num_workers: int = 0, path=None):
        super().__init__(batch_size, num_workers, path)
        self.dataset = MaskedStringDataset(path)



from joblib import Parallel, delayed
from selfies import encoder
from rdkit import Chem

from astrochem_embedding import get_paths


def error_selfies(smi: str):
    try:
        return encoder(smi)
    except:
        return None


def conditional(smi: str) -> bool:
    return all([Chem.MolFromSmiles(smi), "." not in smi])


paths = get_paths()

all_smiles = []

files = [
    paths.get("raw").joinpath("collected_smiles.smi"),
    paths.get("interim").joinpath("kida_isotopologues.smi"),
]

for f in files:
    with open(f) as read_file:
        all_smiles.extend(read_file.readlines())

# get only unique entries
all_smiles = list(set(all_smiles))
all_smiles = [smi for smi in all_smiles if conditional(smi)]

# remove stuff
all_smiles = [smi.strip() for smi in all_smiles]

selfies = Parallel(n_jobs=16)(delayed(error_selfies)(smi) for smi in all_smiles)

with open(paths.get("processed").joinpath("selfies.txt"), "w+") as write_file:
    for selfie in selfies:
        if selfie:
            write_file.write(f"{selfie}\n")



"""Nox sessions."""
import shutil
import sys
from pathlib import Path
from textwrap import dedent

import nox

try:
    from nox_poetry import Session
    from nox_poetry import session
except ImportError:
    message = f"""\
    Nox failed to import the 'nox-poetry' package.

    Please install it using the following command:

    {sys.executable} -m pip install nox-poetry"""
    raise SystemExit(dedent(message))


package = "astrochem_embedding"
python_versions = ["3.9", "3.8", "3.7", "3.6"]
nox.needs_version = ">= 2021.6.6"
nox.options.sessions = (
    "pre-commit",
    "safety",
    "mypy",
    "tests",
    "typeguard",
    "xdoctest",
    "docs-build",
)


def activate_virtualenv_in_precommit_hooks(session: Session) -> None:
    """Activate virtualenv in hooks installed by pre-commit.

    This function patches git hooks installed by pre-commit to activate the
    session's virtual environment. This allows pre-commit to locate hooks in
    that environment when invoked from git.

    Args:
        session: The Session object.
    """
    if session.bin is None:
        return

    virtualenv = session.env.get("VIRTUAL_ENV")
    if virtualenv is None:
        return

    hookdir = Path(".git") / "hooks"
    if not hookdir.is_dir():
        return

    for hook in hookdir.iterdir():
        if hook.name.endswith(".sample") or not hook.is_file():
            continue

        text = hook.read_text()
        bindir = repr(session.bin)[1:-1]  # strip quotes
        if not (
            Path("A") == Path("a") and bindir.lower() in text.lower() or bindir in text
        ):
            continue

        lines = text.splitlines()
        if not (lines[0].startswith("#!") and "python" in lines[0].lower()):
            continue

        header = dedent(
            f"""\
            import os
            os.environ["VIRTUAL_ENV"] = {virtualenv!r}
            os.environ["PATH"] = os.pathsep.join((
                {session.bin!r},
                os.environ.get("PATH", ""),
            ))
            """
        )

        lines.insert(1, header)
        hook.write_text("\n".join(lines))


@session(name="pre-commit", python="3.9")
def precommit(session: Session) -> None:
    """Lint using pre-commit."""
    args = session.posargs or ["run", "--all-files", "--show-diff-on-failure"]
    session.install(
        "black",
        "darglint",
        "flake8",
        "flake8-bandit",
        "flake8-bugbear",
        "flake8-docstrings",
        "flake8-rst-docstrings",
        "pep8-naming",
        "pre-commit",
        "pre-commit-hooks",
        "reorder-python-imports",
    )
    session.run("pre-commit", *args)
    if args and args[0] == "install":
        activate_virtualenv_in_precommit_hooks(session)


@session(python="3.9")
def safety(session: Session) -> None:
    """Scan dependencies for insecure packages."""
    requirements = session.poetry.export_requirements()
    session.install("safety")
    session.run("safety", "check", "--full-report", f"--file={requirements}")


@session(python=python_versions)
def mypy(session: Session) -> None:
    """Type-check using mypy."""
    args = session.posargs or ["src", "tests", "docs/conf.py"]
    session.install(".")
    session.install("mypy", "pytest")
    session.run("mypy", *args)
    if not session.posargs:
        session.run("mypy", f"--python-executable={sys.executable}", "noxfile.py")


@session(python=python_versions)
def tests(session: Session) -> None:
    """Run the test suite."""
    session.install(".")
    session.install("coverage[toml]", "pytest", "pygments")
    try:
        session.run("coverage", "run", "--parallel", "-m", "pytest", *session.posargs)
    finally:
        if session.interactive:
            session.notify("coverage", posargs=[])


@session
def coverage(session: Session) -> None:
    """Produce the coverage report."""
    args = session.posargs or ["report"]

    session.install("coverage[toml]")

    if not session.posargs and any(Path().glob(".coverage.*")):
        session.run("coverage", "combine")

    session.run("coverage", *args)


@session(python=python_versions)
def typeguard(session: Session) -> None:
    """Runtime type checking using Typeguard."""
    session.install(".")
    session.install("pytest", "typeguard", "pygments")
    session.run("pytest", f"--typeguard-packages={package}", *session.posargs)


@session(python=python_versions)
def xdoctest(session: Session) -> None:
    """Run examples with xdoctest."""
    args = session.posargs or ["all"]
    session.install(".")
    session.install("xdoctest[colors]")
    session.run("python", "-m", "xdoctest", package, *args)


@session(name="docs-build", python="3.9")
def docs_build(session: Session) -> None:
    """Build the documentation."""
    args = session.posargs or ["docs", "docs/_build"]
    session.install(".")
    session.install("sphinx", "sphinx-click", "sphinx-rtd-theme")

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-build", *args)


@session(python="3.9")
def docs(session: Session) -> None:
    """Build and serve the documentation with live reloading on file changes."""
    args = session.posargs or ["--open-browser", "docs", "docs/_build"]
    session.install(".")
    session.install("sphinx", "sphinx-autobuild", "sphinx-click", "sphinx-rtd-theme")

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-autobuild", *args)



from functools import cached_property
from typing import Dict, Type, List, Union, Iterable
from pathlib import Path

import numpy as np
import selfies as sf
from ruamel.yaml import YAML

src_path = Path(__file__)
top = src_path.parents[2].absolute()


def get_paths() -> Dict[str, Type[Path]]:
    """
    Retrieve a dictionary containing the absolute paths
    for this project. This provides a simple method for
    traversing and referencing files outside the current
    working directory, particularly for scripts and notebooks.
    """
    paths = {
        "data": top.joinpath("data"),
        "models": top.joinpath("models"),
        "notebooks": top.joinpath("notebooks"),
        "scripts": top.joinpath("scripts"),
    }
    for subfolder in ["raw", "interim", "external", "processed"]:
        paths[subfolder] = paths.get("data").joinpath(subfolder)
    return paths


def get_pretrained_path() -> Type[Path]:
    return src_path.parent.joinpath("models/pretrained")


class Translator(object):
    def __init__(self, alphabet: List[str], max_length: int):
        self.alphabet = alphabet
        self.max_length = max_length

    @property
    def max_length(self) -> int:
        return self._max_length

    @max_length.setter
    def max_length(self, value: int):
        assert value > 0
        self._max_length = value

    @cached_property
    def token_map(self):
        return {s: i for i, s in enumerate(self.alphabet)}

    @property
    def alphabet(self):
        return self._alphabet

    @alphabet.setter
    def alphabet(self, alphabet: List[str]):
        if isinstance(alphabet, list):
            self._alphabet = alphabet

    def __len__(self) -> int:
        return len(self.alphabet)

    def tokenize(self, selfies: str) -> List[int]:
        """
        For backwards compatibility, this tokenizes SELFIES
        for now.
        """
        return self.tokenize_selfies(selfies)

    def tokenize_selfies(self, selfies: str) -> List[int]:
        label, onehot = sf.selfies_to_encoding(selfies, self.token_map, self.max_length)
        return label, onehot

    def tokenize_smiles(self, smiles: str) -> List[int]:
        selfie = sf.encoder(smiles)
        return self.tokenize_selfies(selfie)

    def index_to_character(self, index: int) -> str:
        return self.alphabet[index]

    def indices_to_selfies(self, sentence: Iterable[int]) -> str:
        characters = [
            self.index_to_character(item) for item in sentence if item != "[nop]"
        ]
        return "".join(characters)

    def indices_to_smiles(self, sentence: Iterable[int]) -> str:
        selfie = self.indices_to_selfies(sentence)
        return sf.decoder(selfie)

    @classmethod
    def from_yaml(cls, yaml_path):
        yaml = YAML()
        with open(yaml_path) as read_file:
            data = yaml.load(read_file)
        return cls(data.get("alphabet"), data.get("max_length"))

    @classmethod
    def from_pretrained(cls):
        path = get_pretrained_path().joinpath("translator.yml")
        return cls.from_yaml(path)

    def __repr__(self) -> str:
        return f"Translator with {len(self.alphabet)} tokens, padded to {self.max_length} length."

    def to_yaml(self, yaml_path: str):
        target = Path(yaml_path)
        yaml = YAML()
        with open(target, "w+") as write_file:
            yaml.dump(
                {"alphabet": self.alphabet, "max_length": self.max_length}, write_file
            )



from typing import Union

import pandas as pd
import numpy as np
from rdkit import Chem
from astrochem_ml.smiles.isotopes import isotopologues_from_file, generate_all_isos
from joblib import Parallel, delayed

from astrochem_embedding import get_paths


def convert_inchi_to_smiles(inchi: str) -> Union[None, Chem.Mol]:
    """
    Function to exclude badly InChI from the KIDA molecule
    set. There are a few of them.
    """
    mol = Chem.MolFromInchi(inchi)
    if mol:
        return Chem.MolToSmiles(mol)
    return None


paths = get_paths()

# load in molecules from KIDA
kida_df = pd.read_csv(paths.get("raw").joinpath("kida-molecules_05_Jul_2020.csv"))
# drop rows that don't have data
kida_df.dropna(inplace=True)

smiles = []
# filter out bad InChI and convert to SMILES
for inchi in kida_df["InChI"]:
    smiles.append(convert_inchi_to_smiles(inchi))

smiles = list(filter(lambda x: x is not None, smiles))

# write out the valid KIDA SMILES to file
with open(paths.get("interim").joinpath("kida.smi"), "w+") as write_file:
    for smi in smiles:
        write_file.write(f"{smi}\n")

# exhaustively generate every isotopologues without hydrogen
all_isos = []
for smi in smiles:
    if "c" not in smi:
        all_isos.extend(generate_all_isos(smi, explicit_h=True))
all_isos = list(set(all_isos))

with open(paths.get("interim").joinpath("kida_isotopologues.smi"), "w+") as write_file:
    for iso in all_isos:
        write_file.write(f"{iso}\n")



from astrochem_embedding import utils


def test_pretrained_path():
    path = utils.get_pretrained_path()
    # make sure model weights are there
    assert path.joinpath("VICGAE.pkl").exists()


def test_translator_pretrained():
    t = utils.Translator.from_pretrained()


def test_translator_features():
    t = utils.Translator.from_pretrained()
    assert t.max_length == 886
    assert t.index_to_character(632) == "[nop]"
    # check the tokenization for benzene
    label, onehot = t.tokenize_smiles("c1ccccc1")
    label = list(filter(lambda x: x != 632, label))
    assert label == [373, 250, 373, 250, 373, 250, 528, 242]
